# -*- coding: utf-8 -*-
"""Abstract Mapper."""

# Standard Library Imports
from __future__ import annotations
import abc

__all__ = ["AbstractMapper"]


class AbstractMapper(abc.ABC):
    """Represents an abstract mapper."""
